/*
 * mdns_avahi.c - perform dns service discovery
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <time.h>
#include <avahi-core/core.h>
#include <avahi-core/lookup.h>
#include <avahi-core/log.h>
#include <avahi-common/simple-watch.h>
#include <avahi-common/malloc.h>
#include <avahi-common/error.h>
#include <avahi-common/domain.h>

#include <libubox/uloop.h>
#include <libubus.h>

#include <easy/easy.h>
#include <time.h>

#include "debug.h"
#include "nodes.h"
#include "topologyd.h"
#include "mdns_avahi.h"


struct record_browser_data {
	struct topologyd_private *topo;
	AvahiServer *server;
	AvahiSRecordBrowser *srv_browser;
	AvahiSRecordBrowser *txt_browser;
	AvahiRecord *srv_record;
	AvahiRecord *txt_record;
	const char type[64];
	const char *ap;
	const char domain[256];
	const char *tp;
};

void parse_record(AvahiSRecordBrowser *rr,
		struct record_browser_data *rb,
		AvahiIfIndex interface,
		AvahiProtocol protocol,
		AvahiRecord *record,
		AvahiLookupResultFlags flags);

static AvahiSimplePoll *simple_poll;

//TODO: implement log function properly
void avahi_log_override(AvahiLogLevel level, const char *txt)
{
	return;
}

static struct mdns_service_t *find_service(struct node *n,
		AvahiIfIndex interface, const char *app, const char *trans,
		const char *domain, uint8_t port)
{
	int i;
	struct device_ipv4_t *ipv4 = n->device_ipv4;

	for (i = 0; i < ipv4->mdns_no_services; i++) {
		struct mdns_service_t *s = ipv4->mdns_service[i];

		if (interface == s->ifindex &&
				strcasecmp(s->application_protocol, app) == 0 &&
				strcasecmp(s->transport_protocol, trans) == 0 &&
				strcasecmp(s->domain, domain) == 0 &&
				port == s->port)
			return s;
	}

	return NULL;
}

// will alter 'type'
int parse_protos(char *type, char **ap, char **tp)
{
	char *ptr;

	ptr = strrchr(type, '.');
	if (!ptr)
		return -1;

	// skip _
	*ap = type + 1;
	// separate application proto and transport proto
	*ptr = '\0';
	// skip _
	*tp = ptr + 2;

	return 0;
}

static int parse_txt_records(struct mdns_service_t *s, AvahiStringList *l)
{
	AvahiStringList *n;

	if (s->txt_no_entries >= MDNS_TXT_RECORDS_MAX)
		return -1;

	for (n = l; n; n = n->next) {
		char *ptr;
		int i;

		i = s->txt_no_entries;
		ptr = strstr((char *)n->text, "=");
		if (!ptr)
			continue;

		s->text_record[i] = calloc(1, sizeof(struct text_record_t));

		strncpy(s->text_record[i]->key, (char *)n->text,
				(ptr - (char *)n->text));
		strncpy(s->text_record[i]->value, (ptr + 1),
				(n->size - (ptr - (char *)n->text)));
		s->txt_no_entries++;
	}

	return 0;
}

int parse_mdns_service(struct mdns_service_t *s, const char *ap, const char *tp,
		const char *type, const char *domain, const char *name, uint16_t port,
		uint16_t weight, uint16_t priority, uint8_t interface, uint32_t ttl)
{
	strncpy(s->application_protocol, ap,
			sizeof(s->application_protocol));
	strncpy(s->transport_protocol, tp, sizeof(s->transport_protocol));
	strncpy(s->instance_name, type, sizeof(s->instance_name));
	strncpy(s->domain, domain, sizeof(s->domain));
	strncpy(s->target, name, sizeof(s->target));

	s->port = port;
	s->weight = weight;
	s->priority = priority;
	s->ifindex = interface;
	s->time = time(NULL);
	s->ttl = ttl;

	return 0;
}

static void record_browser_callback(
		AvahiSRecordBrowser *rr,
		AvahiIfIndex interface,
		AvahiProtocol protocol,
		AvahiBrowserEvent event,
		AvahiRecord *record,
		AvahiLookupResultFlags flags,
		void *userdata)
{
	struct record_browser_data *rb = userdata;

	dbg("%s %d event = %d\n", __func__, __LINE__, event);

	switch (event) {
	case AVAHI_BROWSER_REMOVE:
		break;
	case AVAHI_BROWSER_NEW:
		parse_record(rr, rb, interface, protocol, record, flags);
		break;
	case AVAHI_BROWSER_CACHE_EXHAUSTED:
		break;
	case AVAHI_BROWSER_ALL_FOR_NOW:
		break;
	default:
		break;
	}
}

void parse_record(AvahiSRecordBrowser *rr,
		struct record_browser_data *rb,
		AvahiIfIndex interface,
		AvahiProtocol protocol,
		AvahiRecord *record,
		AvahiLookupResultFlags flags)
{
		struct topologyd_private *topo = rb->topo;
		AvahiServer *server = (AvahiServer *) rb->server;
		AvahiRecord *srv_record = rb->srv_record;
		AvahiKey *k;
		const char *domain = rb->domain;

		switch (record->key->type) {
		case AVAHI_DNS_TYPE_A: {
			struct node *n;
			char *ap = NULL, *tp = NULL, type_cpy[64] = {0};
			struct mdns_service_t *s;

			dbg("RECEIVED QUERY %d.%d.%d.%d\n",
					record->data.a.address.address & 0xFF,
					(record->data.a.address.address >> 8) & 0xFF,
					(record->data.a.address.address >> 16) & 0xFF,
					(record->data.a.address.address >> 24) & 0xFF);

			strncpy(type_cpy, rb->type, sizeof(type_cpy));

			if (parse_protos(type_cpy, &ap, &tp))
				break;

			n = node_lookup_ipv4(topo->topo.node_htable,
					record->data.a.address.address);
			if (!n) {
				dbg("No node match found!\n");
				break;
			}

			if (find_service(n, interface, ap, tp, domain,
					srv_record->data.srv.port))
				break;

			n->device_ipv4->mdns_service[n->device_ipv4->mdns_no_services] = s
					= calloc(1, sizeof(*s));
			if (!s)
				break;

			parse_mdns_service(s, ap, tp, rb->type, domain,
					srv_record->data.srv.name, srv_record->data.srv.port,
					srv_record->data.srv.weight, srv_record->data.srv.priority,
					interface, record->ttl);
			parse_txt_records(s, rb->txt_record->data.txt.string_list);

			n->device_ipv4->mdns_no_services++;
			break;
		}
		// TODO: copy paste from ipv4, combine into singular function, untested
		case AVAHI_DNS_TYPE_AAAA: {
			struct node *n;
			char *ap = NULL, *tp = NULL, type_cpy[64] = {0};
			struct mdns_service_t *s;

			strncpy(type_cpy, rb->type, sizeof(type_cpy));

			if (parse_protos(type_cpy, &ap, &tp))
				break;

			n = node_lookup_ipv6(topo->topo.node_htable,
					(uint8_t *)record->data.a.address.address);
			if (!n) {
				dbg("No node match found!\n");
				break;
			}

			if (find_service(n, interface, ap, tp, domain,
					srv_record->data.srv.port))
				break;

			n->device_ipv6->mdns_service[n->device_ipv6->mdns_no_services] = s
					= calloc(1, sizeof(*s));
			if (!s)
				break;

			parse_mdns_service(s, ap, tp, rb->type, domain,
					srv_record->data.srv.name, srv_record->data.srv.port,
					srv_record->data.srv.weight, srv_record->data.srv.priority,
					interface, record->ttl);
			parse_txt_records(s, rb->txt_record->data.txt.string_list);

			n->device_ipv6->mdns_no_services++;
			break;
		}
		case AVAHI_DNS_TYPE_SRV: {
			AvahiSRecordBrowser *record_browser;
			char n[AVAHI_DOMAIN_NAME_MAX];

			if (!record->data.srv.name || protocol)
				break;

			rb->srv_record = avahi_record_ref(record);

			avahi_service_name_join(n, sizeof(n), record->key->name, rb->type,
					domain);
			k = avahi_key_new(record->key->name, AVAHI_DNS_CLASS_IN,
					AVAHI_DNS_TYPE_TXT);

			record_browser = avahi_s_record_browser_new(server, interface,
					protocol, k, 0, record_browser_callback, rb);
			if (!record_browser) {
				avahi_s_record_browser_free(record_browser);
				dbg("%s %d fail initializing SRV record browser!\n",
						__func__, __LINE__);
			}

			rb->srv_browser = rr;
			avahi_key_unref(k);
			break;
		}
		case AVAHI_DNS_TYPE_TXT: {
			AvahiSRecordBrowser *record_browser;

			rb->txt_record = avahi_record_ref(record);

			k = avahi_key_new(srv_record->data.srv.name, AVAHI_DNS_CLASS_IN,
					(protocol ? AVAHI_DNS_TYPE_AAAA : AVAHI_DNS_TYPE_A));

			record_browser = avahi_s_record_browser_new(server, interface,
					protocol, k, 0, record_browser_callback, rb);
			if (!record_browser) {
				avahi_s_record_browser_free(record_browser);
				dbg("%s %d fail initializing TXT record browser!\n",
						__func__, __LINE__);
			}
			rb->txt_browser = rr;
			avahi_key_unref(k);
			break;
		}
		default:
			break;
	}

	if (record->key->type == AVAHI_DNS_TYPE_A ||
			record->key->type == AVAHI_DNS_TYPE_AAAA ) {
		/* TODO: cleaning up browsers probably redundant? */
		avahi_s_record_browser_free(rr);
		avahi_s_record_browser_free(rb->srv_browser);
		avahi_s_record_browser_free(rb->txt_browser);

		avahi_record_unref(srv_record);
		avahi_record_unref(rb->txt_record);
		free(rb);
	}
}

static void service_browser_callback(AvahiSServiceBrowser *b,
		AvahiIfIndex interface, AvahiProtocol protocol, AvahiBrowserEvent event,
		const char *name, const char *type, const char *domain,
		AvahiLookupResultFlags flags, void *userdata)
{
	struct topologyd_private *topo = userdata;
	AvahiServer *server = (AvahiServer *) topo->avahi_serv;

	// TODO: ipv6 support
	if (!name || protocol != AVAHI_PROTO_INET)
		return;

	switch (event) {
	case AVAHI_BROWSER_NEW: {
		struct record_browser_data *tmp = calloc(1, sizeof(*tmp));
		char n[AVAHI_DOMAIN_NAME_MAX];
		AvahiKey *k;
		AvahiSRecordBrowser *record_browser;

		avahi_service_name_join(n, sizeof(n), name, type, domain);
		k = avahi_key_new(n, AVAHI_DNS_CLASS_IN, AVAHI_DNS_TYPE_SRV);
		tmp->topo = topo;
		tmp->server = server;
		strncpy((char *)tmp->type, type, sizeof(tmp->type) - 1);
		strncpy((char *)tmp->domain, domain, sizeof(tmp->domain) - 1);

		record_browser = avahi_s_record_browser_new(server, AVAHI_IF_UNSPEC,
				protocol, k, 0, record_browser_callback, tmp);
		if (!record_browser) {
			avahi_s_record_browser_free(record_browser);
			dbg("%s %d fail initializing SRV record browser!\n");
		}
		avahi_key_unref(k);
		break;
	}
	case AVAHI_BROWSER_REMOVE:
		break;
	case AVAHI_BROWSER_FAILURE:
		fprintf(stderr, "service_browser failed: %s\n",
				avahi_strerror(avahi_server_errno(server)));
		avahi_simple_poll_quit((AvahiSimplePoll *)topo->simple_poll);
		break;
	case AVAHI_BROWSER_CACHE_EXHAUSTED:
		break;
	case AVAHI_BROWSER_ALL_FOR_NOW:
		avahi_simple_poll_quit((AvahiSimplePoll *)topo->simple_poll);
		break;
	}

	avahi_s_service_browser_free(b);
}

static void browse_service_type(struct topologyd_private *topo,
		AvahiServer *server, const char *stype, const char *domain)
{
	AvahiSServiceBrowser *b;

	b = avahi_s_service_browser_new(server, AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC,
			stype, domain, 0, service_browser_callback, topo);
	if (!b) {
		fprintf(stderr, "avahi_service_browser_new() failed: %s\n",
				avahi_strerror(avahi_server_errno(server)));
		avahi_simple_poll_quit(simple_poll);
	}
}

static void clean_text_record(struct text_record_t *t)
{
	dbg("freeing text record -- key: %s value: %s\n", t->key, t->value);
	free(t);
}

static void clean_text_records(struct mdns_service_t *s)
{
	int i;

	dbg("cleaning all text records -- txt_no_entries: %d\n", s->txt_no_entries);

	for (i = s->txt_no_entries-1; i >= 0; i--) {
		clean_text_record(s->text_record[i]);
		s->txt_no_entries--;
	}

	s->txt_no_entries = 0;
}

static void clean_mdns_service(struct mdns_service_t *s)
{
	dbg("cleaning mdns service -- target: %s, application protocol %s, \n"
			"transport protocol %s\n", s->target, s->application_protocol,
			s->transport_protocol);

	clean_text_records(s);
	free(s);
}

void clean_mdns_services(struct device_ipv4_t *d)
{
	int i;

	dbg("cleaning mdns services of device %d.%d.%d.%d\n",
			d->ipv4_addr[0], d->ipv4_addr[1], d->ipv4_addr[2], d->ipv4_addr[3]);

	for (i = d->mdns_no_services - 1; i >= 0; i--) {
		clean_mdns_service(d->mdns_service[i]);
		d->mdns_no_services--;
		d->mdns_service[i] = NULL;
	}

	d->mdns_no_services = 0;
	memset(d->mdns_service, 0, sizeof(d->mdns_service));
}

// Need cleaner way to reset node information
static void node_reset_all(struct hlist_head *table)
{
	struct node *n;
	int i;

	dbg("RESETTING ALL NODES!!\n");

	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&table[i]))
			continue;

		hlist_for_each_entry(n, &table[i], hlist)
			clean_mdns_services(n->device_ipv4);
	}
}

static void browse_callback(AvahiSServiceTypeBrowser *b,
		AVAHI_GCC_UNUSED AvahiIfIndex interface,
		AVAHI_GCC_UNUSED AvahiProtocol protocol, AvahiBrowserEvent event,
		const char *type, const char *domain,
		AVAHI_GCC_UNUSED AvahiLookupResultFlags flags, void *userdata)
{
	struct topologyd_private *topo = userdata;
	AvahiServer *server = (AvahiServer *) topo->avahi_serv;

	switch (event) {
	case AVAHI_BROWSER_NEW:
		browse_service_type(topo, server, type, domain);
		break;

	case AVAHI_BROWSER_REMOVE:
		/* We're dirty and never remove the browser again */
		break;

	case AVAHI_BROWSER_FAILURE:
		fprintf(stderr, "service_type_browser failed: %s\n",
				avahi_strerror(avahi_server_errno(server)));
		avahi_simple_poll_quit((AvahiSimplePoll *)topo->simple_poll);
		break;

	case AVAHI_BROWSER_CACHE_EXHAUSTED:
		break;
	case AVAHI_BROWSER_ALL_FOR_NOW:
		avahi_simple_poll_quit((AvahiSimplePoll *)topo->simple_poll);
		break;
	}
}

static void server_callback(AvahiServer *server, AvahiServerState state,
		AVAHI_GCC_UNUSED void *userdata)
{
	struct topologyd_private *topo = userdata;
	/* Called whenever the client or server state changes */
	if (state == AVAHI_SERVER_FAILURE) {
		fprintf(stderr, "Server connection failure: %s\n",
				avahi_strerror(avahi_server_errno(server)));
		avahi_simple_poll_quit((AvahiSimplePoll *)topo->simple_poll);
	}
}


void mdns_cleanup_poll(struct topologyd_private *topo)
{
	AvahiSimplePoll *simple_poll = (AvahiSimplePoll *) topo->simple_poll;

	if (simple_poll) {
		avahi_simple_poll_free(simple_poll);
		simple_poll = NULL;
	}
}

void mdns_cleanup(struct topologyd_private *topo)
{
	AvahiServer *server = (AvahiServer *) topo->avahi_serv;

	if (server) {
		avahi_server_free(server);
		topo->avahi_serv = NULL;

	}

	mdns_cleanup_poll(topo);
}

int mdns_run(struct topologyd_private *topo)
{
	AvahiServer *server = (AvahiServer *) topo->avahi_serv;
	AvahiSServiceTypeBrowser *sb = NULL;

	node_reset_all(topo->topo.node_htable);
	sb = avahi_s_service_type_browser_new(server, AVAHI_IF_UNSPEC,
			AVAHI_PROTO_UNSPEC, NULL, 0, browse_callback, topo);

	if (!sb) {
		fprintf(stderr, "Failed to create service browser: %s\n",
				avahi_strerror(avahi_server_errno(server)));
		return -1;
	}

	/* Run the main loop */
	avahi_simple_poll_loop((AvahiSimplePoll *)topo->simple_poll);
	avahi_s_service_type_browser_free(sb);

	return 0;
}

int mdns_init(struct topologyd_private *priv)
{
	AvahiServer *server;
	AvahiServerConfig config;
	AvahiSimplePoll *simple_poll;
	int error;

	/* Initialize the pseudo-RNG */
	srand(time(NULL));

	avahi_set_log_function(avahi_log_override);

	/* Allocate main loop object */
	priv->simple_poll = simple_poll  = avahi_simple_poll_new();
	if (!simple_poll) {
		fprintf(stderr, "Failed to create simple poll object.\n");
		return -1;
	}

	/* Do not publish any local records */
	avahi_server_config_init(&config);
	config.publish_hinfo = 0;
	config.publish_addresses = 0;
	config.publish_workstation = 0;
	config.publish_domain = 0;
	config.enable_wide_area = 0;

	/* Allocate a new server */
	priv->avahi_serv = server = avahi_server_new(avahi_simple_poll_get(simple_poll), &config,
			server_callback, priv, &error);

	/* Free the configuration data */
	avahi_server_config_free(&config);

	return 0;
}
