/*
 * topo_ieee1905.c - topology for ieee1905 multiap network
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdbool.h>
#include <errno.h>

#include <libubox/uloop.h>
#include <libubus.h>

#include <easy/easy.h>

#include "debug.h"
#include "topologyd.h"
#include "mdns_avahi.h"
#include "nodes.h"

static bool first_run_flag;

static int topologyd_copy_update_node(struct node *dest, struct node *src)
{
	int j = 0, k = 0;
	if (dest == NULL || src == NULL)
		return -1;

	dest->nbr_num =  src->nbr_num;
	dest->non1905_nbr_num = src->non1905_nbr_num;

	memset(dest->nbr_list, 0, sizeof(dest->nbr_list));
	memset(dest->non1905_nbrlist, 0, sizeof(dest->non1905_nbrlist));

	for (j = 0; j < src->nbr_num; j++) {
		memcpy(dest->nbr_list[j].nbr_macid, src->nbr_list[j].nbr_macid, 6);
	}
	memcpy(dest->non1905_nbrlist, src->non1905_nbrlist,
						6 * src->non1905_nbr_num);

	//copy the non1905 and 1905 neighbor local interface
	memcpy(dest->non1905_nbr_localintf, src->non1905_nbr_localintf, 6);
	memcpy(dest->nbr_localintf, src->nbr_localintf, 6);

	//Here copy the interface name of node
	strncpy(dest->ingress_ifr_name, src->ingress_ifr_name, strlen(src->ingress_ifr_name));

	//Here we copy the interface info
	dest->local_intf_nbr = src->local_intf_nbr;
	for(j = 0; j < src->local_intf_nbr; j++) {
		memcpy(dest->node_intf[j].interface_id, src->node_intf[j].interface_id, 6);
		dest->node_intf[j].media_type = src->node_intf[j].media_type;
		//Here we also need to fill the media specific
		//data for the interface
		dest->node_intf[j].media_specific_data_size = src->node_intf[j].media_specific_data_size;
		if (dest->node_intf[j].media_specific_data_size == 10) {
                                        memcpy(dest->node_intf[j].network_membership,
                                                src->node_intf[j].network_membership, 6);
                                        dest->node_intf[j].role = src->node_intf[j].role;
                                        dest->node_intf[j].ap_channel_band = src->node_intf[j].ap_channel_band;
                                        dest->node_intf[j].freq_index1 = src->node_intf[j].freq_index1;
                                        dest->node_intf[j].freq_index2 = dest->node_intf[j].freq_index2;
                                }
	}

	//Here we copy the L2 neighbor info
	dest->l2_nbr_num = src->l2_nbr_num;	
	for(j = 0; j < src->l2_nbr_num; j++) {
		memcpy(dest->l2_nbr[j].l2_intf_addr, src->l2_nbr[j].l2_intf_addr, 6);
		memcpy(dest->l2_nbr[j].l2_nbr_addr, src->l2_nbr[j].l2_nbr_addr, 6);
		dest->l2_nbr[j].behind_intf_nbr = src->l2_nbr[j].behind_intf_nbr;
		for (k = 0; k < src->l2_nbr[j].behind_intf_nbr; k++) {
			memcpy(dest->l2_nbr[j].behind_intf_addr[k], src->l2_nbr[j].behind_intf_addr[k], 6);
		}
	}

	//Here we copy the Bridge tuple entries
	dest->bridge_tuple_num = src->bridge_tuple_num;
	for(j = 0; j < src->bridge_tuple_num; j++) {
		dest->br_tuple[j].tuple_num = src->br_tuple[j].tuple_num;

		for(k = 0;k < src->br_tuple[j].tuple_num ;k++) {
		        memcpy(dest->br_tuple[j].tuple_macid[k], src->br_tuple[j].tuple_macid[k], 6);
		}
	}

        time(&dest->last_recv);

	return 0;
}

static void topologyd_node_expired(struct uloop_timeout *t)
{
	struct node *p = container_of(t, struct node, validity_timer);
	struct topologyd_private *priv = p->priv;

	//Here we log in the changelog as the node will be deleted
	changelog_copy_node_info(priv, p, 1, 1);

	if (priv) {
		node_del(priv->topo.node_htable, p->hwaddr);
	}
	priv->topo.num_nodes--;
}

int topologyd_node_add(struct topologyd_private *priv, struct node *node)
{
	struct node *p = NULL;

	if (!node || !priv)
		return -1;

	if (priv->topo.num_nodes >= TOPOLOGY_NODE_MAX)
		return -1;

	if (!memcmp(node->hwaddr, priv->ieee1905_macaddr, 6)) {
		p = &priv->selfnode;
		topologyd_copy_update_node(p, node);
		if (first_run_flag == 0) {
			if (!priv->algo_running)
				topologyd_algo_run(priv);
			first_run_flag = 1;
		}
		return 0;
	}

	/* Add node if not present in topology.
	 *
	 * In case the node is present, update the node's values
	 * and the timestamp.
	 */
	p = node_lookup(priv->topo.node_htable, node->hwaddr);
	if (!p) {
		p = node_add(priv->topo.node_htable, node->hwaddr);
		topologyd_copy_update_node(p, node);

		p->validity_timer.cb = topologyd_node_expired;
		p->priv = priv;
		priv->topo.num_nodes++;
		//Add this in changelog events
		changelog_copy_node_info(priv, p, 0, 1); 

	} else {
		//update in changelog
		update_changelog_info(priv, p, node);	
		topologyd_copy_update_node(p, node);
	}

	/* reset validity timer for the node */
	uloop_timeout_set(&p->validity_timer, 180 * 1000);

	return 0;
}

#if 0
//Here we need to check if the node is a stale node and no response
//has been received for more than TOPOLOGY_VALIDITY
void check_node_validity()
{
        int i =0, j=0;
	double diff_t;
	time_t curr_time ;
	time(&curr_time);
        for (i = 0; i < gs_topotree.num_nodes; i++) {
		diff_t = difftime(curr_time, gs_topotree.node[i].last_recv);
                if (diff_t > TOPOLOGY_VALIDITY) {
			//Here we need to delete the node 
			DEBUG("Deleting the node in index [%d] difftime[%f] ", i, diff_t);
			for( j = i; j < gs_topotree.num_nodes -1 ; j++) {
				topologyd_copy_update_node(&(gs_topotree.node[j]), &(gs_topotree.node[j+1]));
			}
			gs_topotree.num_nodes = gs_topotree.num_nodes- 1;
                }
        }
	//Here we may have deleted the node so get self node interfaces
	//updated
	//get_interface_neighbors(NULL);
	topologyd_send_ieee1905_topology_query(priv, NULL);

        return;
}
#endif

/* Build full topology tree using a dumb but effective algo.
 * It uses ieee1905 topology query to probe the neighbor nodes.
 *
 * Assumption: all nodes keep up-to-date list of their immediate neighbors.
 *
 * At any node 'P' in the network -
 *
 * Intial state:
 *		Nbr(P) = {N0,..Nn}  <== immediate neighbors of P.
 *		G{P, N0,..Nn} <== graph of all known nodes at P.
 *
 * Step 1:
 *		for all Ni in G:
 *			Send Topology Request to Ni.
 *
 * Step 2:
 *		Read Topology Response from Ni.
 *		if Nbr(Ni) is not in G:
 *			Insert Nbr(Ni) in G
 *
 * Step 3:
 *		Repeat Step 1-2 until no new Nbr(Ni) can be inserted in G.
 *
 */
void topologyd_algo_run(struct topologyd_private *priv)
{
	struct node *p;
	int i, j;
	//struct linkMetricQueryTLV tlv_metric;

	if (priv->algo_running)
		return;

	dbg("\n======Building Topology======\n");

	priv->algo_running = true;

	if (priv->topo.num_nodes == 0) {
		/* send query to own neighbors */
		p = &priv->selfnode;

		if (hwaddr_is_zero(p->hwaddr))
			topologyd_send_ieee1905_topology_query(priv, NULL);

		dbg("Processing " MACFMT " AL\n", MAC2STR(p->hwaddr));
		dbg("The total nodes are %d", p->nbr_num);

		for (j = 0; j < p->nbr_num; j++) {
			dbg("sending req to " MACFMT "\n", MAC2STR(p->nbr_list[j].nbr_macid));
			topologyd_send_ieee1905_topology_query(priv, p->nbr_list[j].nbr_macid);
			dbg("Processing " MACFMT " AL\n", MAC2STR(p->nbr_list[j].nbr_macid));
		}
	}
	/*here we send the topology query to selfnode each
	 *time the algo runs to keep self node updated*/
	topologyd_send_ieee1905_topology_query(priv, NULL);

	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&priv->topo.node_htable[i]))
			continue;

		hlist_for_each_entry(p, &priv->topo.node_htable[i], hlist) {

			if (hwaddr_is_zero(p->hwaddr))	// TODO: unlikely..
				continue;

			for (j = 0; j < p->nbr_num; j++) {
				dbg("sending req to " MACFMT "\n", MAC2STR(p->nbr_list[j].nbr_macid));
				topologyd_send_ieee1905_topology_query(priv, p->nbr_list[j].nbr_macid);
			}
			//Here we also need to send the topology query to neighbors
			dbg("sending topology req to " MACFMT "\n", MAC2STR(p->hwaddr));
			topologyd_send_ieee1905_topology_query(priv,p->hwaddr);
			//Here we also need to send the query for higher layer request
			dbg("sending higher layer req to " MACFMT "\n", MAC2STR(p->hwaddr));
			topologyd_send_ieee1905_query(priv, p->hwaddr, CMDU_TYPE_HIGHER_LAYER_QUERY, p->ingress_ifr_name, NULL);
			//Here sending the link metric query 
			dbg("sending linkmetric req to " MACFMT "\n", MAC2STR(p->hwaddr));
			//topologyd_getdata_lnmetric_query(&tlv_metric);
			topologyd_send_ieee1905_query(priv, p->hwaddr, CMDU_TYPE_LINK_METRIC_QUERY, p->ingress_ifr_name, NULL);

		}
	}

	priv->algo_running = false;

	if (priv->status == 0)
		uloop_timeout_set(&priv->status_timer, TOPOLOGY_STATUS_INTERVAL*1000);

	/***
	 * gather service information through dns-sd
	 */
	mdns_init(priv);
	mdns_run(priv);
	mdns_cleanup(priv);

	dbg("\n====== Done Building Topology======\n");
}
