/*
 * topologyd.c - topology for multiap network ubus interface
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */
#include <stdbool.h>
#include <pthread.h>
#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubus.h>
#include "nodes.h"

#include <easy/easy.h>

#include "debug.h"
#include "topologyd.h"
#include "config.h"
#include "json_utils.h"

static int signal_pending;
extern const char *ubus_socket;
uint32_t topology_log_max;
static bool g_registrar_freq_band[3];

static void topologyd_sighandler(int sig)
{
	signal_pending = sig;
}

static void ieee1905_info_response_cb(struct ubus_request *req,
					int mtype, struct blob_attr *msg)
{
	struct topologyd_private *p;
	json_object *jroot = NULL;
	char *mac_str = NULL;
	char *str = NULL;
	char *freq_band = NULL;

	if (!msg || !req->priv)
		return;

	p = (struct topologyd_private *)req->priv;

	str = blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jroot = json_tokener_parse(str);
		free(str);
	}

	if (jroot) {
		//FIXME: "ieee1905id" -> "id"
		mac_str = _json_obj_get_string(jroot, "ieee1905id");
		if (mac_str) {
			hwaddr_aton(mac_str, p->ieee1905_macaddr);
			hwaddr_aton(mac_str, p->selfnode.hwaddr);
			free(mac_str);
		}
		/*Here we also need to know if the self node is the registrar node
		 *This is required in the registrar feq band information*/
		freq_band = _json_obj_get_string(jroot, "registrar_freq_band");
		if (freq_band) {

			if((strcasestr(freq_band, "802.11 2.4 GHz")) != NULL) {
				g_registrar_freq_band[0] = 1;
			}
			if((strcasestr(freq_band, "802.11 5 GHz")) != NULL) {
				g_registrar_freq_band[1] = 1;
			}
			if((strcasestr(freq_band, "802.11 60 GHz")) != NULL) {
				g_registrar_freq_band[2] = 1;
			}
			free(freq_band);
		}
		json_safe_put(jroot);
	}
}

static int ieee1905_msg_from_ubus(struct blob_attr *msg, uint16_t *msg_type, uint16_t *msg_mid, uint8_t **tlv, uint32_t *b_len, char * ingress_if_name)
{
	json_object *jroot = NULL;
	char *str;
	char *tlv_str;
	char *if_name;
	int len;
	

	if (!msg)
		return 1;

	str = blobmsg_format_json_indent(msg, true, -1);
	if (str) {
		jroot = json_tokener_parse(str);
		free(str);
	}

	if (jroot) {
		*msg_type = _json_obj_get_int(jroot, "type");
		*msg_mid = _json_obj_get_int(jroot, "mid");
		if_name = _json_obj_get_string(jroot, "ingress");
		if (if_name != NULL) {
			strncpy(ingress_if_name, if_name, strlen(if_name));
			free(if_name);
		}

		tlv_str = _json_obj_get_string(jroot, "tlvs");
		if (tlv_str == NULL)
			return 1;
		len = strlen(tlv_str);
		//dbg(" The TLV string is %s LENGTH Id %d", tlv_str, len);
		*b_len = (len/2);		
		//dbg("The binary length of the tlv is %d", *b_len);
		*tlv = (uint8_t *) malloc((*b_len) * sizeof(uint8_t));
		if (*tlv == NULL) {
			err("No Memory \n");
			if (tlv_str != NULL)
				free(tlv_str);
			return 1;
		}
		strtob(tlv_str, len, *tlv);
		if (tlv_str != NULL)
			free(tlv_str);
		json_safe_put(jroot);
	}

	return 0;
}

int topologyd_get_ieee1905_macaddr(struct topologyd_private *p)
{
	struct blob_buf bb = { 0 };
	int ret = 0;

	blob_buf_init(&bb, 0);
	ret = ubus_invoke(p->ctx, p->ieee1905, "info", bb.head,
				ieee1905_info_response_cb, p, 1000);
	if (ret)
		err("Failed to get ieee1905 info\n");

	blob_buf_free(&bb);

	return ret;
}

int topologyd_send_ieee1905_topology_query(struct topologyd_private *p,
						uint8_t *dest)
{
	char dest_macstr[18] = { 0 };
	struct blob_buf bb = { 0 };
	int ret = 0;


	blob_buf_init(&bb, 0);
	if (dest) {
		hwaddr_ntoa(dest, dest_macstr);
		blobmsg_add_string(&bb, "dst_macaddr", dest_macstr);
	}

	ret = ubus_invoke(p->ctx, p->ieee1905, "query_topology", bb.head,
					NULL, NULL, 1000);
	if (ret)
		err("Failed to send ieee1905 topology query\n");


	blob_buf_free(&bb);

	return ret;
}

void topologyd_event_handler(struct ubus_context *ctx,
			struct ubus_event_handler *ev, const char *type,
			struct blob_attr *msg)
{
	int ret = 0;
	struct cmdu_cstruct *cstruct = NULL;
	uint16_t msg_type = 0;
	uint16_t msg_mid = 0;
	uint8_t *tlv = NULL;
	uint32_t len = 0;
	char itfr_name[256] = {0}; 

	struct topologyd_private *priv =
			container_of(ev, struct topologyd_private, ev);

	if (!msg)
		return;

	//Calling the function to get the args from json msg
	ret = ieee1905_msg_from_ubus(msg, &msg_type, &msg_mid, &tlv, &len, itfr_name);
	if (ret != 0) {
		err("ieee1905_msg_from_ubus error \n");
		if (tlv != NULL)
			free(tlv);
		return;
	}

	dbg("Inside topology event handler %hu %hu  %s", msg_type, msg_mid, itfr_name);

	if (!((msg_type == CMDU_TYPE_TOPOLOGY_RESPONSE) || (msg_type == CMDU_TYPE_TOPOLOGY_NOTIFICATION) || (msg_type == CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE) || (msg_type == CMDU_TYPE_HIGHER_LAYER_RESPONSE) || (msg_type == CMDU_TYPE_LINK_METRIC_RESPONSE))) {
		if (tlv != NULL)
                	free(tlv);
		return;
	}


	//cstruct = ieee1905_build_cmdu(msg_type, msg_mid, 0, tlv, len);
	cstruct = ieee1905_build_cmdu(msg_type, msg_mid, 0, tlv);

	dbg("After Event handler \n");

	if (cstruct == NULL)
		return;

	switch(cstruct->message_type) {
		case CMDU_TYPE_TOPOLOGY_RESPONSE:
		{
			dbg("Received topology response  message \n");
			topologyd_process_topology_response(cstruct, priv, itfr_name);
		}
		break;
		case CMDU_TYPE_TOPOLOGY_NOTIFICATION:
		{
			dbg("Received topology notification message \n");
			/*Here we recieved the topology notification
			 * we muct run the algo to update the topology*/ 
			if (!priv->algo_running)
                		topologyd_algo_run(priv);
		}
		case CMDU_TYPE_HIGHER_LAYER_RESPONSE:
		{
			dbg("Received higher layer response message \n");
			topologyd_process_higherlayer_response(cstruct, priv);
		}	
		break;
		case CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE:
		{
			dbg("Received ap autoconfiguration response message \n");
			topologyd_process_autoconfiguration_response(cstruct, priv);
		}	
		break;
		case CMDU_TYPE_LINK_METRIC_RESPONSE:
		{
			dbg("Received link metric response");
			topologyd_process_linkmetric_response(cstruct, priv);

		}
		break;
		default:
			dbg("Event handler received default message \n");
		break;
	}
	ieee1905_free_cmdu(cstruct);
	if (tlv != NULL)
		free(tlv);
	return;
}


void topologyd_process_topology_response(struct cmdu_cstruct *cstruct, struct topologyd_private *priv, char *itfr_name)
{

	int i = 0, j = 0, k = 0, l =0, cnt= 0;
	struct node n = { 0 };
	uint8_t *p;

	if(cstruct == NULL || priv == NULL)
		return;

	dbg("Inside topologyd process response \n");
	while (NULL != (p = cstruct->tlv_cstruct_list[i])) {

		dbg("  tlv type is %x \n", *p);
		switch (*p) {
		case TLV_TYPE_DEVICE_INFORMATION_TYPE:
		{
			struct deviceInformationTypeTLV *tlv =
				(struct deviceInformationTypeTLV *)p;

			memcpy(&n.hwaddr, &tlv->al_mac_address, 6);

			/*Here as we fill the device related information
			 * also fill the interface info from which 
			 * we get this topology response
			 */
			if (itfr_name != NULL)
				strncpy(n.ingress_ifr_name, itfr_name, strlen(itfr_name));
			//Here we need to fill the interface fields
			n.local_intf_nbr = tlv->local_interfaces_nr;
			for (j = 0; j < n.local_intf_nbr; j++) {
				memcpy(n.node_intf[j].interface_id, &tlv->local_interfaces[j].mac_address, 6);
				n.node_intf[j].media_type = tlv->local_interfaces[j].media_type;
			
				//Here we also need to fill the media specific
				//data for the interface
				n.node_intf[j].media_specific_data_size = tlv->local_interfaces[j].media_specific_data_size;
			//	dbg("Print of the octave %o hexa %x decimal %d \n", n.node_intf[j].media_specific_data_size, n.node_intf[j].media_specific_data_size, n.node_intf[j].media_specific_data_size);
				if (tlv->local_interfaces[j].media_specific_data_size == 10) {
					memcpy(n.node_intf[j].network_membership, 
						&tlv->local_interfaces[j].media_specific_data.ieee80211.network_membership, 6);
					n.node_intf[j].role = tlv->local_interfaces[j].media_specific_data.ieee80211.role;
					n.node_intf[j].ap_channel_band = tlv->local_interfaces[j].media_specific_data.ieee80211.ap_channel_band;
					n.node_intf[j].freq_index1 = tlv->local_interfaces[j].media_specific_data.ieee80211.ap_channel_center_frequency_index_1;
					n.node_intf[j].freq_index2 = tlv->local_interfaces[j].media_specific_data.ieee80211.ap_channel_center_frequency_index_2;

				}

			}
		}
		break;
		case TLV_TYPE_NEIGHBOR_DEVICE_LIST:
		{
			struct neighborDeviceListTLV  *tlv =
				(struct neighborDeviceListTLV *)p;
			memcpy(n.nbr_localintf, &tlv->local_mac_address, 6);
			n.nbr_num = tlv->neighbors_nr;
			for (j = 0; j < n.nbr_num; j++) {
				memcpy(n.nbr_list[j].nbr_macid, &tlv->neighbors[j].mac_address, 6);
			}
		}
		break;
		case TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST:
		{
			struct non1905NeighborDeviceListTLV *tlv =
				(struct non1905NeighborDeviceListTLV *)p;
			memcpy(n.non1905_nbr_localintf, &tlv->local_mac_address,6);
			n.non1905_nbr_num = tlv->non_1905_neighbors_nr;
			for (j = 0; j < n.non1905_nbr_num; j++) {
				memcpy(n.non1905_nbrlist[j], &tlv->non_1905_neighbors[j].mac_address, 6);
			}
		}
		break;
		case TLV_TYPE_L2_NEIGHBOR_DEVICE:
		{
			struct l2NeighborDeviceTLV *tlv = 
				(struct l2NeighborDeviceTLV *)p;
			cnt = 0;
			dbg("Inside topologyd response l2 message intf_num[%d] \n", tlv->local_interfaces_nr);

                        for (j = 0; j < tlv->local_interfaces_nr; j++) {

				for (k = 0; k < tlv->local_interfaces[j].l2_neighbors_nr; k++) {

				dbg("Inside topologyd response L2 message intf_num[%d] nbr_num[%d] behind_mac_num[%d] \n", tlv->local_interfaces_nr, tlv->local_interfaces[j].l2_neighbors_nr,tlv->local_interfaces[j].l2_neighbors[k].behind_mac_addresses_nr);
					memcpy(n.l2_nbr[cnt].l2_intf_addr, &tlv->local_interfaces[j].local_mac_address, 6);
					memcpy(n.l2_nbr[cnt].l2_nbr_addr, &tlv->local_interfaces[j].l2_neighbors[k].l2_neighbor_mac_address, 6);
					n.l2_nbr[cnt].behind_intf_nbr = tlv->local_interfaces[j].l2_neighbors[k].behind_mac_addresses_nr;

					for (l = 0; l < n.l2_nbr[cnt].behind_intf_nbr; l++) {

						memcpy(n.l2_nbr[cnt].behind_intf_addr[l], &tlv->local_interfaces[j].l2_neighbors[k].behind_mac_addresses[l], 6);
					}
					cnt++;
				}
                        }
			n.l2_nbr_num = cnt;
		}
		break;
		case TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES:
		{
			struct deviceBridgingCapabilityTLV *tlv =
                                (struct deviceBridgingCapabilityTLV *)p;

			n.bridge_tuple_num = tlv->bridging_tuples_nr;
                        dbg(" bridging tuple number [%d] \n", tlv->bridging_tuples_nr);

			for (j = 0; j < tlv->bridging_tuples_nr; j++) {
				n.br_tuple[j].tuple_num = tlv->bridging_tuples[j].bridging_tuple_macs_nr;

                                for (k = 0; k < tlv->bridging_tuples[j].bridging_tuple_macs_nr; k++) {
					 dbg(" bridge_num[%d] br_mac_num[%d] \n", tlv->bridging_tuples_nr, tlv->bridging_tuples[j].bridging_tuple_macs_nr);
                                        memcpy(n.br_tuple[j].tuple_macid[k], &tlv->bridging_tuples[j].bridging_tuple_macs[k].mac_address, 6);
				}
			}
		}
		break;
		default :
			dbg (" received a default tlv \n");
		break;
		}
		i++;
	}

	if (n.nbr_num || n.non1905_nbr_num) {
		dbg("Updating Node" MACFMT " from topology response\n",
						MAC2STR(n.hwaddr));
		topologyd_node_add(priv, &n);
	}
}

void topologyd_process_higherlayer_response(struct cmdu_cstruct *cstruct, struct topologyd_private *priv)
{
	int i = 0, j = 0, res = 0, k = 0, cnt = 0;
	struct node *n = NULL;
	uint8_t al_mac_address[6];
	uint8_t *p;
	struct x1905ProfileVersionTLV *profile_ver = NULL;
	struct deviceIdentificationTypeTLV *identification = NULL;
	struct controlUrlTypeTLV *control_url = NULL;
	struct ipv4TypeTLV *ipv4 = NULL;
	struct ipv6TypeTLV *ipv6 = NULL;
	bool al_mac_address_is_present = 0;

	if(cstruct == NULL || priv == NULL )
		return;

	dbg("Inside Higher Layer response \n");
	while (NULL != (p = cstruct->tlv_cstruct_list[i])) {

		switch (*p) {
		case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
		{
			struct alMacAddressTypeTLV *tlv = (struct alMacAddressTypeTLV *)p;
			memcpy(al_mac_address, tlv->al_mac_address, 6);
			al_mac_address_is_present = 1;
		}
		break;
		case TLV_TYPE_1905_PROFILE_VERSION:
		{
			profile_ver = (struct x1905ProfileVersionTLV *)p;
		}
		case TLV_TYPE_DEVICE_IDENTIFICATION:
		{
			identification = (struct deviceIdentificationTypeTLV *)p;
		}
		break;
		case TLV_TYPE_CONTROL_URL:
		{
			control_url = (struct controlUrlTypeTLV *)p;
		}
		break;
		case TLV_TYPE_IPV4:
		{
			ipv4 = (struct ipv4TypeTLV *)p;
			dbg(" Inside ipv4 tlv \n");
		}
		break;
		case TLV_TYPE_IPV6:
		{
			ipv6 = (struct ipv6TypeTLV *)p;
			dbg(" Inside ipv6 tlv \n");
		}
		break;

		}
		i++;
	}

	if (al_mac_address_is_present) {
		dbg("Updating Node with the higher layer response message\n");
		//Here we 1st compare that the node is self node
		res = memcmp (al_mac_address, priv->selfnode.hwaddr, 6);
		if (res == 0) {
			dbg(" The HL resp corresponds to the self node we are ignoring this details\n");
			return;
		}
		n = node_lookup(priv->topo.node_htable, al_mac_address);
		//If we find the node in our devices 
		if( n != NULL) {
			//getting the node specific params
			if (profile_ver != NULL)
				n->version = profile_ver->profile; 
			if (identification != NULL) {
				strncpy(n->friendly_name, identification->friendly_name, 64);
				strncpy(n->manufacturer_name, identification->manufacturer_name, 64);
				strncpy(n->manufacturer_model, identification->manufacturer_model, 64);
			}
			if(control_url != NULL)
				strncpy(n->control_url, control_url->url, strlen(control_url->url));

			//getting ipv4 specific params
			cnt = 0;
			if (ipv4 != NULL) {
				
                        	for (j = 0; j < ipv4->ipv4_interfaces_nr; j++) {

                                	for (k = 0; k < ipv4->ipv4_interfaces[j].ipv4_nr; k++) {

                                		dbg("Inside topologyd hhl message intf_num[%d] ipv4_num[%d] \n", 
							ipv4->ipv4_interfaces_nr, ipv4->ipv4_interfaces[j].ipv4_nr);
                                        	memcpy(n->device_ipv4[cnt].mac_address, &ipv4->ipv4_interfaces[j].mac_address, 6);
                                        	memcpy(n->device_ipv4[cnt].ipv4_addr, &ipv4->ipv4_interfaces[j].ipv4[k].ipv4_address, 4);
                                        	n->device_ipv4[cnt].ipv4_addr_type = ipv4->ipv4_interfaces[j].ipv4[k].type;
                                        	memcpy(n->device_ipv4[cnt].dhcp_addr, &ipv4->ipv4_interfaces[j].ipv4[k].ipv4_dhcp_server, 4);
                                        	cnt++;
                                	}
                        	}
                        	n->ipv4_intf_nbr = cnt;
			}
			cnt = 0;	

			//getting the ipv6 specific params
			if(ipv6 != NULL) {

				for (j = 0; j < ipv6->ipv6_interfaces_nr; j++) {

                                	for (k = 0; k < ipv6->ipv6_interfaces[j].ipv6_nr; k++) {

                                        	dbg("Inside topologyd hhl message intf_num[%d] ipv6_num[%d] \n",
                                                        ipv6->ipv6_interfaces_nr, ipv6->ipv6_interfaces[j].ipv6_nr);
                                        	memcpy(n->device_ipv6[cnt].mac_address, &ipv6->ipv6_interfaces[j].mac_address, 6);
                                        	memcpy(n->device_ipv6[cnt].ipv6_addr, &ipv6->ipv6_interfaces[j].ipv6[k].ipv6_address, 16);
                                        	n->device_ipv6[cnt].ipv6_addr_type = ipv6->ipv6_interfaces[j].ipv6[k].type;
                                        	memcpy(n->device_ipv6[cnt].ipv6addr_origin, &ipv6->ipv6_interfaces[j].ipv6[k].ipv6_address_origin, 16);
                                        	cnt++;
                                	}
                        	}
				n->ipv6_intf_nbr = cnt;
			}
		}
	}
}

void topologyd_process_autoconfiguration_response(struct cmdu_cstruct *cstruct, struct topologyd_private *priv)
{
	int i = 0;
	uint8_t *p;
	uint32_t supported_role_is_present = 0;
	uint32_t supported_freq_band_is_present = 0;
	uint8_t supported_freq_band = IEEE80211_FREQUENCY_BAND_2_4_GHZ;

	if(cstruct == NULL || priv == NULL)
		return;

	dbg("Inside autoconfiguration process response \n");
	while (NULL != (p = cstruct->tlv_cstruct_list[i])) {

		switch (*p) {
		case TLV_TYPE_SUPPORTED_ROLE:
		{
			supported_role_is_present = 1;
		}
		break;
		case TLV_TYPE_SUPPORTED_FREQ_BAND:
		{
			struct supportedFreqBandTLV *tlv =
				(struct supportedFreqBandTLV *)p;
			supported_freq_band_is_present = 1;
			supported_freq_band = tlv->freq_band;
		}
		break;
		}
                i++;
        }

        if (supported_role_is_present == 1 && supported_freq_band_is_present == 1) {
		dbg("Updating registrar freq band \n");
		//Here the boolean value of the supported freq band is changed to 1
		if((supported_freq_band >= 0) && (supported_freq_band <= 2))
			g_registrar_freq_band[supported_freq_band] = 1;
		
	}
}

static int topologyd_refresh(struct ubus_context *ctx,
				struct ubus_object *obj,
				struct ubus_request_data *req,
				const char *method,
				struct blob_attr *msg)
{
	struct topologyd_private *priv =
			container_of(obj, struct topologyd_private, obj);

	if (!priv->algo_running)
		topologyd_algo_run(priv);

	return 0;
}


void topologyd_process_linkmetric_response(struct cmdu_cstruct *cstruct, struct topologyd_private *priv)
{
	int i = 0;
	uint8_t *p;

	if(cstruct == NULL || priv == NULL )
		return;

	dbg("Inside Link metric response \n");
	while (NULL != (p = cstruct->tlv_cstruct_list[i])) {

		switch (*p) {
		case TLV_TYPE_TRANSMITTER_LINK_METRIC:
		{
			dbg (" Inside transmitter link metric\n");
			topology_add_txrx_data(p , priv);
		}
		break;
		case TLV_TYPE_RECEIVER_LINK_METRIC:
		{
			dbg (" Inside receiver link metric \n");
			topology_add_txrx_data(p , priv);
		}
		break;
		case TLV_TYPE_VENDOR_SPECIFIC:
		{
			//Nothing to do with the vendor tlv now
		}
		break;
		}
		i++;
	}

}

void topology_add_txrx_data (uint8_t *metric, struct topologyd_private *priv)
{
	uint8_t fr_al_mac_id[6];
	uint8_t to_al_mac_id[6];
	uint32_t i, j, res; 
	struct node *n = NULL;

	if (metric == NULL)
		return;

	dbg("Updating Node with the Link layer response message\n");
	//Here we 1st compare that the node is self node

	if(*metric == TLV_TYPE_TRANSMITTER_LINK_METRIC) {
		dbg (" Inside the transmitter link metric tlv\n");
		struct transmitterLinkMetricTLV *ptr;

		ptr = (struct transmitterLinkMetricTLV *)metric;

		memcpy(fr_al_mac_id, ptr->local_al_address,6);
		memcpy(to_al_mac_id, ptr->neighbor_al_address, 6);

		res = memcmp (fr_al_mac_id, priv->selfnode.hwaddr, 6);
		if (res == 0) {
			dbg(" The Link Layer resp corresponds to the self node we are ignoring this details\n");
			return;
		}

		n = node_lookup(priv->topo.node_htable, fr_al_mac_id);

		//If we find the node in our devices
		if( n != NULL) {
			//getting the node specific params then we need to search the neighbor for which this metric is given
			for (i = 0; i < n->nbr_num; i++) {
				res = memcmp (n->nbr_list[i].nbr_macid, to_al_mac_id, 6);
				if(res != 0)
					continue;
				n->nbr_list[i].metric_num = ptr->transmitter_link_metrics_nr;
				dbg (" Inside the transmitter link metric tlv  ln_metric_num %d\n", ptr->transmitter_link_metrics_nr);
				for (j = 0; j < ptr->transmitter_link_metrics_nr; j++) {
					memcpy(n->nbr_list[i].ln_metric[j].nbr_ifr_addr, ptr->transmitter_link_metrics[j].neighbor_interface_address, 6);
					n->nbr_list[i].ln_metric[j].ieee802dot1br = ptr->transmitter_link_metrics[j].bridge_flag;
					n->nbr_list[i].ln_metric[j].tx_packet_errors = ptr->transmitter_link_metrics[j].packet_errors;
					n->nbr_list[i].ln_metric[j].tx_packet_ok = ptr->transmitter_link_metrics[j].transmitted_packets;
					n->nbr_list[i].ln_metric[j].tx_max_xput = ptr->transmitter_link_metrics[j].mac_throughput_capacity;
					n->nbr_list[i].ln_metric[j].tx_link_availability = ptr->transmitter_link_metrics[j].link_availability;
					n->nbr_list[i].ln_metric[j].tx_phy_rate = ptr->transmitter_link_metrics[j].phy_rate;
				}
			}
		}
	}
	else if(*metric == TLV_TYPE_RECEIVER_LINK_METRIC) {

                struct receiverLinkMetricTLV *p;
		dbg (" Inside the receiver link metric tlv\n");

                p = (struct receiverLinkMetricTLV *)metric;

		memcpy(fr_al_mac_id, p->local_al_address,6);
                memcpy(to_al_mac_id, p->neighbor_al_address, 6);
		
                res = memcmp (fr_al_mac_id, priv->selfnode.hwaddr, 6);
                if (res == 0) {
                        dbg(" The Link Layer resp corresponds to the self node we are ignoring this details\n");
                        return;
                }

                n = node_lookup(priv->topo.node_htable, fr_al_mac_id);
                //If we find the node in our devices
                if( n != NULL) {
                        //getting the node specific params then we need to search the neighbor for which this metric is given
                        for (i = 0; i < n->nbr_num; i++) {
                                res = memcmp (n->nbr_list[i].nbr_macid, to_al_mac_id, 6);
                                if(res != 0)
                                        continue;
				dbg (" Inside the receiver link metric  ln_num is %d\n", p->receiver_link_metrics_nr);
                                for (j = 0; j < p->receiver_link_metrics_nr; j++) {
                                        n->nbr_list[i].ln_metric[j].rx_packet_errors = p->receiver_link_metrics[j].packet_errors;
                                        n->nbr_list[i].ln_metric[j].rx_packet_ok = p->receiver_link_metrics[j].packets_received;
                                        n->nbr_list[i].ln_metric[j].rx_rssi = p->receiver_link_metrics[j].rssi;
                                }
                        }
                }
	}
}



void topologyd_dump_node(struct blob_buf *bb, struct node *p, int is_self)
{
	char mac_str[18] = { 0x0 };
	char freq_str[50] = { 0x0 };
	char behind_mac_str[1024] = { 0x0 };
	void *nbr_array, *non1905_array = NULL, *interfaces = NULL, *table = NULL, *l2_neighbor = NULL;
	void *bridge_tuple = NULL, *ipv4_param = NULL, *ipv6_param = NULL, *link_array = NULL;
	void *bridge_tuple_mac = NULL;
	void *table1 = NULL;
	int  j, k, flag = 0, i = 0;
	char addr[16] = {0};
	char addr_v6[40] = {0};

	hwaddr_ntoa(p->hwaddr, mac_str);
        blobmsg_add_string(bb, "ieee1905_macaddr", mac_str);
        blobmsg_add_u32(bb, "num_neighbors", p->nbr_num);

        nbr_array = blobmsg_open_array(bb, "neighbors");
        for (j = 0; j < p->nbr_num; j++) {
		table1 = blobmsg_open_table(bb, NULL);
                hwaddr_ntoa(p->nbr_list[j].nbr_macid, mac_str);
                blobmsg_add_string(bb, "nbr_mac_id", mac_str);
		if (is_self == 0) {
			link_array = blobmsg_open_array(bb, "link_metrics");
			for (k = 0; k < p->nbr_list[j].metric_num; k++) {
				table = blobmsg_open_table(bb, NULL);
				hwaddr_ntoa(p->nbr_list[j].ln_metric[k].nbr_ifr_addr, mac_str);
				blobmsg_add_string(bb, "nbr_intf_macid", mac_str);
				blobmsg_add_u32(bb,"ieee_bridge",p->nbr_list[j].ln_metric[k].ieee802dot1br);
				blobmsg_add_u32(bb,"tx_packet_error",p->nbr_list[j].ln_metric[k].tx_packet_errors);
				blobmsg_add_u32(bb,"tx_packet_ok",p->nbr_list[j].ln_metric[k].tx_packet_ok);
				blobmsg_add_u32(bb,"mac_thrput_capacity",p->nbr_list[j].ln_metric[k].tx_max_xput);
				blobmsg_add_u32(bb,"link_availability",p->nbr_list[j].ln_metric[k].tx_link_availability);
				blobmsg_add_u32(bb,"phy_rate",p->nbr_list[j].ln_metric[k].tx_phy_rate);
				blobmsg_add_u32(bb,"rx_packet_error",p->nbr_list[j].ln_metric[k].rx_packet_errors);
				blobmsg_add_u32(bb,"rx_packet_ok",p->nbr_list[j].ln_metric[k].rx_packet_ok);
				blobmsg_add_u32(bb,"rssi",p->nbr_list[j].ln_metric[k].rx_rssi);
				blobmsg_close_table(bb, table);
			}
			blobmsg_close_array(bb, link_array);
		}
		blobmsg_close_table(bb, table1);
        }
        blobmsg_close_array(bb, nbr_array);

        blobmsg_add_u32(bb, "num_non1905_neighbors", p->non1905_nbr_num);
        non1905_array = blobmsg_open_array(bb, "non1905_neighbors");
        for (j = 0; j < p->non1905_nbr_num; j++) {
                hwaddr_ntoa(p->non1905_nbrlist[j], mac_str);
                blobmsg_add_string(bb, NULL, mac_str);
        }
	blobmsg_close_array(bb, non1905_array);

	if(is_self == 1)
		return;

	/*Here we dump the registar information*/
	for (i = 0; i < 3; i++)
	{
		if (g_registrar_freq_band[i] == 1) {
			if(flag == 1)
				strncat(freq_str, ",", 1);
			strncat(freq_str, get_registra_freq_string(i), 15);
			flag = 1;
		}
	}
	blobmsg_add_string(bb, "registrar_freq_band", freq_str);


	/*Here we are dumping the local interfaces of the 1905 neighbor
	 * and non1905 neighbor of the device*/
	if (p->non1905_nbr_num != 0) {
		hwaddr_ntoa(p->non1905_nbr_localintf, mac_str);
		blobmsg_add_string(bb, "non1905_nbr_localintf", mac_str);
	}

	if (p->nbr_num != 0) {
		hwaddr_ntoa(p->nbr_localintf, mac_str);
		blobmsg_add_string(bb, "1905_nbr_localintf", mac_str);
	}

	//Here we show the Device Local Interface Info
	blobmsg_add_u32(bb, "local_interface_nbr", p->local_intf_nbr);
	interfaces = blobmsg_open_array(bb, "interfaces");
        for (j = 0; j < p->local_intf_nbr; j++) {
		table = blobmsg_open_table(bb, NULL);
                hwaddr_ntoa(p->node_intf[j].interface_id, mac_str);
                blobmsg_add_string(bb, "interface_id", mac_str);
		blobmsg_add_string(bb,"media_type",get_interface_type_string(p->node_intf[j].media_type));
		if(p->node_intf[j].media_specific_data_size == 10) {
			hwaddr_ntoa(p->node_intf[j].network_membership, mac_str);
                	blobmsg_add_string(bb, "network_membership", mac_str);
			blobmsg_add_string(bb,"role",get_role_type_string(p->node_intf[j].role));
			blobmsg_add_u32(bb,"ap_channel_band",p->node_intf[j].ap_channel_band);
			blobmsg_add_u32(bb,"freq_index1",p->node_intf[j].freq_index1);
			blobmsg_add_u32(bb,"freq_index2",p->node_intf[j].freq_index2);
		}
		blobmsg_close_table(bb, table);
        }
        blobmsg_close_array(bb, interfaces);

	//Here we show the Device L2 Nodes Info
	blobmsg_add_u32(bb, "l2_neighbor_num", p->l2_nbr_num);
	l2_neighbor = blobmsg_open_array(bb, "l2_neighbor");
        for (j = 0; j < p->l2_nbr_num; j++) {
		table = blobmsg_open_table(bb, NULL);
		hwaddr_ntoa(p->l2_nbr[j].l2_intf_addr, mac_str);
		blobmsg_add_string(bb, "local_intf_id", mac_str);
		//TODO add more information for l2
		hwaddr_ntoa(p->l2_nbr[j].l2_nbr_addr, mac_str);
                blobmsg_add_string(bb, "local_nbr_id", mac_str);
                for (k = 0; k < p->l2_nbr[j].behind_intf_nbr; k++) {
			hwaddr_ntoa(p->l2_nbr[j].behind_intf_addr[k], mac_str);
			if ( k != 0)
				strcat(behind_mac_str, ",");
			strcat(behind_mac_str, mac_str);
                }
		blobmsg_add_string(bb, "behind_mac_id", behind_mac_str);
		blobmsg_close_table(bb, table);
	}
	blobmsg_close_array(bb, l2_neighbor);

	//Here we show the Bridging tuple enteries
	blobmsg_add_u32(bb, "bridging_tuple_num", p->bridge_tuple_num);
        bridge_tuple = blobmsg_open_array(bb, "bridge_tuple");
        for (j = 0; j < p->bridge_tuple_num; j++) {
                table = blobmsg_open_table(bb, NULL);
		bridge_tuple_mac = blobmsg_open_array(bb, "br_mac");
		for (k = 0; k < p->br_tuple[j].tuple_num; k++) {
			hwaddr_ntoa(p->br_tuple[j].tuple_macid[k], mac_str);
			blobmsg_add_string(bb, "bridge_tuple_mac_id", mac_str);
		}
		blobmsg_close_array(bb, bridge_tuple_mac);
		blobmsg_close_table(bb, table);
        }
        blobmsg_close_array(bb, bridge_tuple);

	//Here we show the higher layer message data
	blobmsg_add_string(bb, "version", get_version(p->version));
	blobmsg_add_string(bb, "friendly_name", p->friendly_name);
	blobmsg_add_string(bb, "manufacturer_name", p->manufacturer_name);
	blobmsg_add_string(bb, "manufacturer_model", p->manufacturer_model);
	blobmsg_add_string(bb, "control_url", p->control_url);

	//Here we show the ipv4 message structure
	blobmsg_add_u32(bb, "ipv4_itfr_num", p->ipv4_intf_nbr);
	ipv4_param = blobmsg_open_array(bb, "ipv4_params");

	for (j = 0; j < p->ipv4_intf_nbr; j++) {
		int k;
		void *service_arr = NULL;

		table = blobmsg_open_table(bb, NULL);
		hwaddr_ntoa(p->device_ipv4[j].mac_address, mac_str);
		blobmsg_add_string(bb, "mac_address", mac_str);
		blobmsg_add_string(bb, "ipv4_addr_type",get_ipv4_type(p->device_ipv4[j].ipv4_addr_type));

		// ipv4 address
		sprintf(addr, "%d.%d.%d.%d", p->device_ipv4[j].ipv4_addr[0],
				p->device_ipv4[j].ipv4_addr[1], p->device_ipv4[j].ipv4_addr[2],
				p->device_ipv4[j].ipv4_addr[3]);
		addr[15] = '\0';
		blobmsg_add_string(bb, "ipv4_address", addr);
		memset(addr, 0, 16);

		// dhcp address
		if (p->device_ipv4[j].ipv4_addr_type == IPV4_TYPE_DHCP) {
			sprintf(addr, "%d.%d.%d.%d", p->device_ipv4[j].dhcp_addr[0],
				p->device_ipv4[j].dhcp_addr[1],
				p->device_ipv4[j].dhcp_addr[2],
				p->device_ipv4[j].dhcp_addr[3]);
			addr[15] = '\0';
			blobmsg_add_string(bb, "dhcp_addr", addr);
		}

		blobmsg_add_u16(bb, "mdns_no_services", p->device_ipv4->mdns_no_services);
		service_arr = blobmsg_open_array(bb, "mdns_services");

		for (k = 0; k < p->device_ipv4->mdns_no_services; k++) {
			int l;
			struct mdns_service_t *service;
			void *service_obj, *txt_arr;
			char date[20];

			service = p->device_ipv4->mdns_service[k];

			service_obj = blobmsg_open_table(bb, "NULL");
			blobmsg_add_string(bb, "application_protocol", service->application_protocol);
			blobmsg_add_string(bb, "transport_protocol", service->transport_protocol);
			blobmsg_add_string(bb, "domain", service->domain);
			blobmsg_add_u16(bb, "port", service->port);
			blobmsg_add_string(bb, "target", service->target);
			blobmsg_add_u16(bb, "txt_no_entries", service->txt_no_entries);

			txt_arr = blobmsg_open_array(bb, "text_records");
			for (l = 0; l < service->txt_no_entries; l++) {
				void *txt_table;

				txt_table = blobmsg_open_table(bb, "NULL");
				blobmsg_add_string(bb, "key", service->text_record[l]->key);
				blobmsg_add_string(bb, "value", service->text_record[l]->value);
				blobmsg_close_table(bb, txt_table);
			}
			blobmsg_close_table(bb, txt_arr);

			strftime(date, 20, "%Y-%m-%d %H:%M:%S", localtime(&service->time));
			blobmsg_add_string(bb, "last_update", date);
			blobmsg_add_u16(bb, "time_to_live", service->ttl);
			blobmsg_add_u16(bb, "priority", service->priority);
			blobmsg_add_u16(bb, "weight", service->weight);
			blobmsg_close_table(bb, service_obj);
		}
		blobmsg_close_array(bb, service_arr);
		blobmsg_close_table(bb, table);
	}
	blobmsg_close_array(bb, ipv4_param);

	//Here we show the ipv6 message structure
	blobmsg_add_u32(bb, "ipv6_itfr_num", p->ipv6_intf_nbr);
        ipv6_param = blobmsg_open_array(bb, "ipv6_params");
        for (j = 0; j < p->ipv6_intf_nbr; j++) {
                table = blobmsg_open_table(bb, NULL);
                hwaddr_ntoa(p->device_ipv6[j].mac_address, mac_str);
                blobmsg_add_string(bb, "mac_address", mac_str);
                blobmsg_add_string(bb, "ipv6_addr_type",get_ipv6_type(p->device_ipv6[j].ipv6_addr_type));
                //For ipv6 address
		sprintf(addr_v6, "%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x",
                                p->device_ipv6[j].ipv6_addr[0], p->device_ipv6[j].ipv6_addr[1], p->device_ipv6[j].ipv6_addr[2],
                                p->device_ipv6[j].ipv6_addr[3], p->device_ipv6[j].ipv6_addr[4], p->device_ipv6[j].ipv6_addr[5],
                                p->device_ipv6[j].ipv6_addr[6], p->device_ipv6[j].ipv6_addr[7], p->device_ipv6[j].ipv6_addr[8],
                                p->device_ipv6[j].ipv6_addr[9], p->device_ipv6[j].ipv6_addr[10], p->device_ipv6[j].ipv6_addr[11],
                                p->device_ipv6[j].ipv6_addr[12], p->device_ipv6[j].ipv6_addr[13], p->device_ipv6[j].ipv6_addr[14],
                                p->device_ipv6[j].ipv6_addr[15]);
                addr_v6[39] = '\0';
                blobmsg_add_string(bb, "ipv6_address", addr_v6);
		// origin address
		if (p->device_ipv6[j].ipv6_addr_type == IPV6_TYPE_SLAAC) {
			memset(addr_v6, 0, 40);
                        sprintf(addr_v6, "%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x",
                                        p->device_ipv6[j].ipv6addr_origin[0], p->device_ipv6[j].ipv6addr_origin[1],
				       	p->device_ipv6[j].ipv6addr_origin[2], p->device_ipv6[j].ipv6addr_origin[3],
                                        p->device_ipv6[j].ipv6addr_origin[4], p->device_ipv6[j].ipv6addr_origin[5],
                                        p->device_ipv6[j].ipv6addr_origin[6], p->device_ipv6[j].ipv6addr_origin[7], 
					p->device_ipv6[j].ipv6addr_origin[8], p->device_ipv6[j].ipv6addr_origin[9],
                                        p->device_ipv6[j].ipv6addr_origin[10], p->device_ipv6[j].ipv6addr_origin[11],
                                        p->device_ipv6[j].ipv6addr_origin[12], p->device_ipv6[j].ipv6addr_origin[13],
				       	p->device_ipv6[j].ipv6addr_origin[14],p->device_ipv6[j].ipv6addr_origin[15]);
                        addr_v6[39] = '\0';
                        blobmsg_add_string(bb, "origin", addr_v6);
                }
                blobmsg_close_table(bb, table);
        }
        blobmsg_close_array(bb, ipv6_param);

}

void topologyd_dump_node_nbr(struct topologyd_private *priv, struct blob_buf *b)
{
	char mac_str[18] = { 0x0 };
        void *nodes_array;
        struct node *p;
        int i;

        /* Add details of own node first */
        p = &priv->selfnode;

	hwaddr_ntoa(p->hwaddr, mac_str);
        blobmsg_add_string(b, "ieee1905_macaddr", mac_str);

	nodes_array = blobmsg_open_array(b, "neighbors");
        for (i = 0; i < NODE_HTABLE_SIZE; i++) {
                if (hlist_empty(&priv->topo.node_htable[i]))
                        continue;

                hlist_for_each_entry(p, &priv->topo.node_htable[i], hlist) {

                        if (hwaddr_is_zero(p->hwaddr))  // TODO: unlikely..
                                continue;

			hwaddr_ntoa(p->hwaddr, mac_str);
			blobmsg_add_string(b, NULL, mac_str);
                }
        }
        blobmsg_close_array(b, nodes_array);
}

void topologyd_dump_tree(struct topologyd_private *priv, struct blob_buf *bb)
{
	void *nodes_array, *table;
	struct node *p;
	int i;

	/* Add details of own node first */
	p = &priv->selfnode;

	table = blobmsg_open_table(bb, "self");
	topologyd_dump_node(bb, p, 1);
	blobmsg_close_table(bb, table);

	nodes_array = blobmsg_open_array(bb, "nodes");
	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&priv->topo.node_htable[i]))
			continue;

		hlist_for_each_entry(p, &priv->topo.node_htable[i], hlist) {

			if (hwaddr_is_zero(p->hwaddr))	// TODO: unlikely..
				continue;

			table = blobmsg_open_table(bb, NULL);

			topologyd_dump_node(bb, p, 0);
			blobmsg_close_table(bb, table);
		}
	}
	blobmsg_close_array(bb, nodes_array);
}

static int topologyd_dump(struct ubus_context *ctx,
				struct ubus_object *obj,
				struct ubus_request_data *req,
				const char *method,
				struct blob_attr *msg)
{
	struct blob_buf bb;
	struct topologyd_private *priv =
			container_of(obj, struct topologyd_private, obj);

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	topologyd_dump_tree(priv, &bb);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

static int topologyd_nodes(struct ubus_context *ctx,
                                struct ubus_object *obj,
                                struct ubus_request_data *req,
                                const char *method,
                                struct blob_attr *msg)
{
        struct blob_buf bb;
        struct topologyd_private *priv =
                        container_of(obj, struct topologyd_private, obj);

        memset(&bb, 0, sizeof(struct blob_buf));
        blob_buf_init(&bb, 0);

        topologyd_dump_node_nbr(priv, &bb);

        ubus_send_reply(ctx, req, bb.head);
        blob_buf_free(&bb);

        return 0;
}


void topologyd_print_changelog(struct blob_buf *bb, struct topologyd_private *priv, uint32_t start, uint32_t end) 
{
	struct topology_changelog *log;
	char rpt_ifmacstr[18] = { 0 };
        char rpt_macstr[18] = { 0 };
        char nbr_macstr[18] = { 0 };
        void *t;
        int i;
        struct tm *info;
        time_t tmp_t;
	char str_tm[20];
	
	dbg("changelog start value is [%d] end value is [%d]", start, end); 
	if (bb == NULL || priv == NULL)
		return;

	for (i = start ; i <= end ; i++) {
		log = &priv->topo.changelog[i];
		t = blobmsg_open_table(bb, NULL);

		hwaddr_ntoa(log->rpt_macaddr, rpt_macstr);
		blobmsg_add_string(bb, "reporter", rpt_macstr);

		hwaddr_ntoa(log->rpt_ifmacaddr, rpt_ifmacstr);
		blobmsg_add_string(bb, "reporter_interface", rpt_ifmacstr);

		hwaddr_ntoa(log->nbr_macaddr, nbr_macstr);
		blobmsg_add_string(bb, "neighbor", nbr_macstr);

		blobmsg_add_string(bb, "is1905_neighbor",
			log->is1905_nbr ? "IEEE1905" : "Non-IEEE1905");

		blobmsg_add_string(bb, "event_type",
			log->type ? "delete" : "add");

		tmp_t = log->timestamp;
		info = localtime(&tmp_t);
		strftime(str_tm, sizeof str_tm, "%Y-%m-%dT%H:%M:%S", info);
		blobmsg_add_string(bb, "timestamp", str_tm);
		blobmsg_close_table(bb, t);
	}
}
void topologyd_dump_changelog(struct topologyd_private *priv, struct blob_buf *bb)
{
	void *a;
	blobmsg_add_u32(bb, "num_changelog", priv->topo.num_changelog);

	a = blobmsg_open_array(bb, "changelog");
	if(priv->topo.changelog_front == -1) {
		blobmsg_close_array(bb, a);
		return;
	}
	if(priv->topo.changelog_front > priv->topo.changelog_rear) {
		topologyd_print_changelog(bb, priv, priv->topo.changelog_front, topology_log_max-1);
		topologyd_print_changelog(bb, priv, 0, priv->topo.changelog_rear);
	}
	else {
		topologyd_print_changelog(bb, priv, 0, priv->topo.changelog_rear);
	}

	blobmsg_close_array(bb, a);
}

static int topologyd_changelog(struct ubus_context *ctx,
				struct ubus_object *obj,
				struct ubus_request_data *req,
				const char *method,
				struct blob_attr *msg)
{
	struct blob_buf bb;
	struct topologyd_private *priv =
			container_of(obj, struct topologyd_private, obj);

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	topologyd_dump_changelog(priv, &bb);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

static int topologyd_status(struct ubus_context *ctx,
				struct ubus_object *obj,
				struct ubus_request_data *req,
				const char *method,
				struct blob_attr *msg)
{
	struct blob_buf bb;
	struct topologyd_private *priv =
			container_of(obj, struct topologyd_private, obj);

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "status", priv->status == AVAILABLE ?
					"available" : "incomplete");

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

static void topologyd_periodic_refresh(struct uloop_timeout *t)
{
	struct topologyd_private *priv =
		container_of(t, struct topologyd_private, refresh_timer);

	dbg("Refresh topology...\n");

	topologyd_algo_run(priv);

	if (priv->config.refresh_int)
		uloop_timeout_set(&priv->refresh_timer, priv->config.refresh_int * 1000);
}

static void topologyd_change_status(struct uloop_timeout *t)
{
        struct topologyd_private *priv =
                container_of(t, struct topologyd_private, status_timer);

        dbg("Change the status of topology...\n");
	priv->status = 1;

}


static void topologyd_start_heartbeat(struct uloop_timeout *t)
{
	struct topologyd_private *priv =
			container_of(t, struct topologyd_private, heartbeat);

	if (!signal_pending)
		uloop_timeout_set(&priv->heartbeat, 1 * 1000);

	switch(signal_pending) {
	case SIGTERM:
	case SIGINT:
		//topologyd_exit(priv);
		exit(0);
		break;
	case SIGHUP:
		//topologyd_reload(priv);
		break;
	default:
		break;
	}
}

static int topologyd_run(struct topologyd_private *priv)
{
	int ret;

	ret = topologyd_get_ieee1905_macaddr(priv);
	ret |= topologyd_send_ieee1905_topology_query(priv, NULL);

	if (ret) {
		err("Failed to run topologyd...\n");
		return -1;
	}

	
	topologyd_start_heartbeat(&priv->heartbeat);
	topologyd_periodic_refresh(&priv->refresh_timer);

	//initalize the changelog values
	priv->topo.changelog_front = priv->topo.changelog_rear = -1;

	return 0;
}

static uint32_t topologyd_get_ieee1905_object(struct topologyd_private *priv)
{
	uint32_t id;
	int status;

	status = ubus_lookup_id(priv->ctx, IEEE1905_OBJECT, &id);
	if (status != UBUS_STATUS_OK) {
		err("object '%s' not present!\n", IEEE1905_OBJECT);
		return OBJECT_INVALID;
	}

	return id;
}

int topologyd_publish_object(struct topologyd_private *priv, const char *objname)
{
	struct ubus_object *obj;
	struct ubus_object_type *obj_type;
	struct ubus_method *obj_methods;
	struct ubus_method m[5] = {
		UBUS_METHOD_NOARG("refresh", topologyd_refresh),
		UBUS_METHOD_NOARG("status", topologyd_status),
		UBUS_METHOD_NOARG("dump", topologyd_dump),
		UBUS_METHOD_NOARG("changelog", topologyd_changelog),
		UBUS_METHOD_NOARG("nodes", topologyd_nodes),
	};
	int num_methods = ARRAY_SIZE(m);
	int ret;


	obj = &priv->obj;
	memset(obj, 0, sizeof(*obj));

	obj_type = calloc(1, sizeof(struct ubus_object_type));
	if (!obj_type)
		return -1;

	obj_methods = calloc(num_methods, sizeof(struct ubus_method));
	if (!obj_methods) {
		free(obj_type);
		return -1;
	}

	obj->name = objname;
	memcpy(obj_methods, m, num_methods * sizeof(struct ubus_method));
	obj->methods = obj_methods;
	obj->n_methods = num_methods;

	obj_type->name = obj->name;
	obj_type->n_methods = obj->n_methods;
	obj_type->methods = obj->methods;
	obj->type = obj_type;

	ret = ubus_add_object(priv->ctx, obj);
	if (ret) {
		fprintf(stderr, "Failed to add 'obj' err = %s\n",
							ubus_strerror(ret));
		free(obj_methods);
		free(obj_type);

		return ret;
	}

	fprintf(stderr, "Added object '%s'\n", objname);

	return 0;
}

int topologyd_start(void)
{
	struct topologyd_private *priv;

	signal_pending = 0;
	set_sighandler(SIGINT, topologyd_sighandler);
	set_sighandler(SIGTERM, topologyd_sighandler);
	set_sighandler(SIGHUP, topologyd_sighandler);
	set_sighandler(SIGPIPE, SIG_IGN);

	priv = calloc(1, sizeof(*priv));
	if (!priv)
		return -ENOMEM;

	info("Starting topologyd...");

	uloop_init();
	priv->ctx = ubus_connect(ubus_socket);
	if (!priv->ctx) {
		err("Failed to connect to ubus\n");
		free(priv);
		return -1;
	}

	ubus_add_uloop(priv->ctx);

	for (;;) {
		priv->ieee1905 = topologyd_get_ieee1905_object(priv);
		if (priv->ieee1905 != OBJECT_INVALID)
			break;

		sleep(2);
	}
	//topologyd_load_extras();	//TODO

	topologyd_config_defaults(&priv->config);
	topologyd_config_user(&priv->config);

	priv->ev.cb = topologyd_event_handler;
	ubus_register_event_handler(priv->ctx, &priv->ev, IEEE1905_EVENTS);

	priv->refresh_timer.cb = topologyd_periodic_refresh;
	priv->heartbeat.cb = topologyd_start_heartbeat;
	priv->topo.changelog = (struct topology_changelog*) malloc 
			((priv->config.maxlog)*sizeof(struct topology_changelog));
	if (priv->topo.changelog == NULL)
		goto out_and_exit;

	topology_log_max = priv->config.maxlog;
	dbg("Dynamic array size %d",priv->config.maxlog);

	priv->status = 0;
	priv->status_timer.cb = topologyd_change_status;

	if (!priv->config.enabled) {
		info("topologyd not enabled.\n");
	} else {
		topologyd_publish_object(priv, TOPOLOGY_OBJECT);
		topologyd_run(priv);
	}

	uloop_run();

	ubus_unregister_event_handler(priv->ctx, &priv->ev);
	//topologyd_remove_object(priv->ctx);	//TODO
out_and_exit:
	ubus_free(priv->ctx);
	uloop_done();
	free(priv);

	return 0;
}

void topologyd_stop(struct topologyd_private *priv)
{
	if (!priv)
		exit(0);

	ubus_unregister_event_handler(priv->ctx, &priv->ev);
	//topologyd_remove_object(priv->ctx);
	ubus_free(priv->ctx);
	uloop_done();
	free(priv);
}

// Check if the changelog queue is full
int is_full(struct topology *t) 
{
	if ((t->changelog_front == t->changelog_rear + 1) || (t->changelog_front == 0 && t->changelog_rear == topology_log_max - 1)) 
		return 1;
	return 0;
}

// Adding an entry in changelog
void enqueue_changelog(struct topology *t, struct topology_changelog *c) 
{
	dbg("Inside enqueue_changelog \n");
	struct topology_changelog *elem;	
	if (is_full(t)) {
		dbg("Changelog queue is full deleting the front node \n");
		t->changelog_front = (t->changelog_front + 1) % topology_log_max;
		t->num_changelog--;
	}
	
	if (t->changelog_front == -1) 
		t->changelog_front = 0;
	t->changelog_rear = (t->changelog_rear + 1) % topology_log_max;
	elem =&(t->changelog[t->changelog_rear]);
	memset(elem, 0, sizeof(struct topology_changelog ));
	memcpy(elem->rpt_macaddr, c->rpt_macaddr, 6);
	memcpy(elem->rpt_ifmacaddr, c->rpt_ifmacaddr, 6);
	memcpy(elem->nbr_macaddr, c->nbr_macaddr, 6);
	elem->is1905_nbr = c->is1905_nbr;
	elem->type = c->type;
	elem->timestamp = time(NULL); 
	//Maintain the number of elements in changelog
	t->num_changelog++;
}

void changelog_copy_node_info(struct topologyd_private *t, struct node *p, int32_t event, bool is1905)
{
	dbg("Inside changelog_copy_node_info \n");
	struct topology_changelog elem;

	if (t == NULL || p == NULL)
		return; 

	if (event == 0) {
		memcpy(elem.rpt_macaddr, p->hwaddr, 6);
		memcpy(elem.rpt_ifmacaddr, p->nbr_localintf, 6);
	}
	else {
		memcpy(elem.rpt_macaddr, t->ieee1905_macaddr, 6);
		memcpy(elem.rpt_ifmacaddr, t->selfnode.nbr_localintf, 6);
	}
	
	memcpy(elem.rpt_ifmacaddr, p->nbr_localintf, 6);
	memcpy(elem.nbr_macaddr, p->hwaddr, 6);
	elem.is1905_nbr = is1905;
	elem.type = event;
	enqueue_changelog(&(t->topo), &elem);
}

void update_changelog_info(struct topologyd_private *priv, struct node *dest, struct node *src)
{
	dbg("Inside update_changelog_info \n");
	int i = 0, j = 0, k = 0, found = 0;
	struct topology_changelog elem;

	if (dest == NULL || src == NULL || priv == NULL)
                return;

        //Check the update in neighbor node 
        if (dest->nbr_num !=  src->nbr_num) {
		//Here a new neighbor is added or deleted
		if (dest->nbr_num > src->nbr_num) {
			dbg("Inside update_changelog_info deleting neighbor node\n");
			// neighbor node is deleted
			for( i = 0; i< dest->nbr_num ; i++) {
				found = 0;
				for(j = 0; j < src->nbr_num; j++) {
					if (memcmp(dest->nbr_list[i].nbr_macid, src->nbr_list[j].nbr_macid, 6) == 0){
						found = 1;
						break;
					}
				}
				if (found == 0) {
					memcpy(elem.nbr_macaddr, dest->nbr_list[i].nbr_macid, 6);
					memcpy(elem.rpt_macaddr, dest->hwaddr, 6);
					memcpy(elem.rpt_ifmacaddr, dest->nbr_localintf, 6);
					elem.is1905_nbr = 1;
					elem.type = 1;
					enqueue_changelog(&(priv->topo), &elem);
				}
			}
		}
		else {
			dbg("Inside update_changelog_info adding neighbor node\n");
			//neighbor node is added
			k = src->nbr_num - dest->nbr_num;
			for( i = 0; i< k ; i++) {
				memcpy(elem.nbr_macaddr, src->nbr_list[i].nbr_macid, 6);
				memcpy(elem.rpt_macaddr, src->hwaddr, 6);
				memcpy(elem.rpt_ifmacaddr, src->nbr_localintf, 6);
				elem.is1905_nbr = 1;
				elem.type = 0;
				enqueue_changelog(&(priv->topo), &elem);
			}
		}
	}

	//Check the update in non neighbor node
        if (dest->non1905_nbr_num != src->non1905_nbr_num) {
		//Here a new non ieee1905 neighbor is added or deleted
		if (dest->non1905_nbr_num > src->non1905_nbr_num) {
                        // non1905 neighbor node is deleted
			dbg("Inside update_changelog_info deleting non neighbor node\n");
			for( i = 0; i< dest->non1905_nbr_num ; i++) {
				found = 0;
				for(j = 0; j < src->non1905_nbr_num; j++) {
					if (memcmp(dest->non1905_nbrlist[i], src->non1905_nbrlist[j], 6) == 0){
						found = 1;
						break;
					}
				}
				if (found == 0) {
					memcpy(elem.nbr_macaddr, dest->non1905_nbrlist[i], 6);
					memcpy(elem.rpt_macaddr, dest->hwaddr, 6);
					memcpy(elem.rpt_ifmacaddr, dest->non1905_nbr_localintf, 6);
					elem.is1905_nbr = 0;
					elem.type = 1;
					enqueue_changelog(&(priv->topo), &elem);
				}
			}
		}
                else {
                        //non1905 neighbor node is added
			dbg("Inside update_changelog_info adding non neighbor node\n");
                        k = src->non1905_nbr_num - dest->non1905_nbr_num;
                        for( i = 0; i< k ; i++) {
                                memcpy(elem.nbr_macaddr, src->non1905_nbrlist[i], 6);
                                memcpy(elem.rpt_macaddr, src->hwaddr, 6);
                                memcpy(elem.rpt_ifmacaddr, src->non1905_nbr_localintf, 6);
                                elem.is1905_nbr = 0;
                                elem.type = 0;
                                enqueue_changelog(&(priv->topo), &elem);
                        }
                }
	}
	return;
}

#if 0
void topologyd_getdata_lnmetric_query(struct linkMetricQueryTLV *tlv)
{
	if (tlv == NULL)
		return;
	memset(tlv, 0, sizeof(struct linkMetricQueryTLV));
	tlv->tlv_type = TLV_TYPE_LINK_METRIC_QUERY;
	tlv->destination = LINK_METRIC_QUERY_TLV_ALL_NEIGHBORS;
	memset(&tlv->specific_neighbor, 0, 6);
	tlv->link_metrics_type = LINK_METRIC_QUERY_TLV_BOTH_TX_AND_RX_LINK_METRICS;
	return;
}

#endif

int topologyd_send_ieee1905_query(struct topologyd_private *p,
                uint8_t *dest, uint32_t type, char * egress, uint8_t *data)
{
        char dest_macstr[18] = { 0 };
        struct blob_buf bb = { 0 };
        int ret = 0;

        blob_buf_init(&bb, 0);
        if (dest) {
		//fill the destination of the message
                hwaddr_ntoa(dest, dest_macstr);
                blobmsg_add_string(&bb, "dst_macaddr", dest_macstr);
        }

	if (type) {
		//fill the type of message
		blobmsg_add_u32(&bb, "type", type);
	}

	if (egress) {
		blobmsg_add_string(&bb, "egress", egress);
	}

	if (data) {
		//fill the tlv of the message by converting binary to string
		

	}		

        ret = ubus_invoke(p->ctx, p->ieee1905, "send_query", bb.head,
                                        NULL, NULL, 1000);
        if (ret)
                err("Failed to send ieee1905 query\n");


        blob_buf_free(&bb);

        return ret;
}

const char *get_interface_type_string(uint16_t if_type)
{

        switch (if_type) {
        case MEDIA_TYPE_IEEE_802_3U_FAST_ETHERNET:
                return "IEEE 802.3u";
        case MEDIA_TYPE_IEEE_802_3AB_GIGABIT_ETHERNET:
                return "IEEE 802.3ab";
        case MEDIA_TYPE_IEEE_802_11B_2_4_GHZ:
                return "IEEE 802.11b";
        case MEDIA_TYPE_IEEE_802_11G_2_4_GHZ:
                return "IEEE 802.11g";
        case MEDIA_TYPE_IEEE_802_11A_5_GHZ:
                return "IEEE 802.11a";
        case MEDIA_TYPE_IEEE_802_11N_2_4_GHZ:
                return "IEEE 802.11n 2.4";
        case MEDIA_TYPE_IEEE_802_11N_5_GHZ:
                return "IEEE 802.11n 5.0";
        case MEDIA_TYPE_IEEE_802_11AC_5_GHZ:
                return "IEEE 802.11ac";
        case MEDIA_TYPE_IEEE_802_11AD_60_GHZ:
                return "IEEE 802.11ad";
        case MEDIA_TYPE_IEEE_802_11AF_GHZ:
                return "IEEE 802.11af";
        case MEDIA_TYPE_IEEE_1901_WAVELET:
                return "IEEE 1901 Wavelet";
        case MEDIA_TYPE_IEEE_1901_FFT:
                return "IEEE 1901 FFT";
        case MEDIA_TYPE_MOCA_V1_1:
                return "MoCAv1.1";
        case MEDIA_TYPE_UNKNOWN:
                return "Generic PHY";
        }
        return "UNKNOWN";
}

const char *get_role_type_string(uint16_t role_type)
{
	switch(role_type) {
		case IEEE80211_SPECIFIC_INFO_ROLE_AP:
     			return"AP";
		case IEEE80211_SPECIFIC_INFO_ROLE_NON_AP_NON_PCP_STA:
			return"non-AP/non-PCP STA";
		case IEEE80211_SPECIFIC_INFO_ROLE_WIFI_P2P_CLIENT:
			return"Wi-Fi P2P Client";
		case IEEE80211_SPECIFIC_INFO_ROLE_WIFI_P2P_GROUP_OWNER:
			return"Wi-Fi P2P Group Owner";
		case IEEE80211_SPECIFIC_INFO_ROLE_AD_PCP:
			return"802.11adPCP";
	}
	return "UNKNOWN";
}

const char *get_registra_freq_string(uint32_t band)
{
	switch(band) {
		case 0:
			return"802.11 2.4 GHz";
		case 1:
			return"802.11 5 GHz";
		case 2:
			return"802.11 60 GHz";
	}
	return "UNKNOWN";
}

const char *get_version(uint8_t ver)
{
        switch(ver) {
                case PROFILE_1905_1:
                        return"1905.1";
                case PROFILE_1905_1A:
                        return"1905.1a";
        }
        return "UNKNOWN";
}

const char *get_ipv4_type(uint8_t ipv4_type)
{
	switch(ipv4_type) {
                case IPV4_TYPE_UNKNOWN:
                        return"IPV4_UNKNOWN";
                case IPV4_TYPE_DHCP:
                        return"IPV4_DHCP";
		case IPV4_TYPE_STATIC:
			return"IPV4_STATIC_IP";
		case IPV4_TYPE_AUTOIP:
			return"IPV4_AUTO_IP";
        }
        return "UNKNOWN";
}

const char *get_ipv6_type(uint8_t ipv6_type)
{
        switch(ipv6_type) {
                case IPV6_TYPE_UNKNOWN:
                        return"IPV6_UNKNOWN";
                case IPV6_TYPE_DHCP:
                        return"IPV6_DHCP";
                case IPV6_TYPE_STATIC:
                        return"IPV6_STATIC_IP";
                case IPV6_TYPE_SLAAC:
                        return"IPV6_SLAAC";
        }
        return "UNKNOWN";
}
